import os

from invoke import Collection, Context, task

from . import utils


CONFIG = {
    "docker": {
        # Sensitive
        "repo": None,
        "user": "AWS",
        "token": None,
        "arg": None,
        # Non-sensitive
        "name": "webserver",
        "context": ".",
        "file": "Dockerfile",
        "tag": "test",
    }
}


# Docker deployment workflow:


@task()
def login(
    ctx: Context,
    repo: str | None = None,
    user: str | None = None,
    token: str | None = None,
) -> None:
    """Login to Docker remote registry (AWS ECR or Dockerhub)."""
    user = user or ctx.docker.user
    if user == utils.UserType.AWS:
        _aws_login(ctx, repo)
    else:
        _dockerhub_login(ctx, user, token)


def _aws_login(ctx: Context, repo: str | None) -> None:
    repo = repo or ctx.docker.repo
    assert repo is not None, "Missing remote AWS ECR URL."
    cmd = "aws ecr get-login-password | "
    cmd += f"docker login --username AWS --password-stdin {repo}"
    ctx.run(cmd)


def _dockerhub_login(ctx: Context, user: str, token: str | None) -> None:
    token = token or ctx.docker.token
    assert token is not None, "Missing remote Dockerhub token."
    ctx.run(f"docker login --username {user} --password {token}")


@task()
def configure(ctx: Context, builder: str = "multi-platform-builder") -> None:
    """Create and provision buildx builder for multi-platform."""
    ctx.run(f"docker buildx create --name {builder} --use")
    ctx.run("docker buildx inspect --bootstrap")


@task()
def build(
    ctx: Context,
    name: str | None = None,
    context: str | None = None,
    file: str | None = None,
    arg: str | None = None,
) -> None:
    """Build Docker image locally."""
    name = name or ctx.docker.name
    context, file = _get_build_context(ctx, context, file)
    cmd = f"docker build --pull --file {file} --tag {name} "
    cmd += _get_build_arg_fragment(ctx, arg)
    cmd += f"{context}"
    ctx.run(cmd)


@task(
    help={
        "tag-only": "Whether to not tag the Docker image as latest",
        "dry-run": "Whether to only tag the Docker image and to not push",
    }
)
def push(
    ctx: Context,
    repo: str | None = None,
    name: str | None = None,
    tag: str | None = None,
    tag_only: bool = False,
    dry_run: bool = False,
) -> None:
    """Push Docker image to a remote registry."""
    repo = repo or ctx.docker.repo
    name = name or ctx.docker.name
    tag = tag or ctx.docker.tag
    tag_list = [tag] if tag_only else [tag, "latest"]
    for t in tag_list:
        target_name = f"{repo}/{name}:{t}"
        ctx.run(f"docker tag {name} {target_name}")
        if not dry_run:
            ctx.run(f"docker push {target_name}")


@task(help={"tag-only": "Whether to not tag the Docker image as latest"})
def buildx(
    ctx: Context,
    repo: str | None = None,
    name: str | None = None,
    context: str | None = None,
    file: str | None = None,
    arg: str | None = None,
    platform: str = "linux/arm64,linux/amd64",
    tag: str | None = None,
    tag_only: bool = False,
) -> None:
    """Build and push x-platform Docker image to a remote registry."""
    # :Warning:
    # Without using `--push` option to push the image, expect the error:
    # ```No output specified with docker-container driver.
    # Build result will only remain in the build cache.```
    # To push image into registry use `--push`
    # or to load image into docker use `--load`
    # In addition: `--load` option does not work for multiple platforms
    repo = repo or ctx.docker.repo
    name = name or ctx.docker.name
    context, file = _get_build_context(ctx, context, file)
    tag = tag or ctx.docker.tag
    tag_list = [tag] if tag_only else [tag, "latest"]
    for t in tag_list:
        target_name = f"{repo}/{name}:{t}"
        cmd = f"docker buildx build --pull --file {file} --tag {target_name} "
        cmd += _get_build_arg_fragment(ctx, arg)
        cmd += f"--platform {platform} --push "
        cmd += f"{context}"
        ctx.run(cmd)


def _get_build_context(ctx: Context, context: str | None, file: str | None) -> tuple[str, str]:
    context = context or ctx.docker.context
    file = file or f"{context}/{ctx.docker.file}"
    utils.check_path(context)
    utils.check_path(file)
    return (context, file)


def _get_build_arg_fragment(ctx: Context, arg: str | None) -> str:
    arg = arg or ctx.docker.arg
    cmd = ""
    if arg is not None:
        assert arg in os.environ, f"Missing environment variable {arg}."
        cmd += f"--build-arg {arg}={os.environ[arg]} "
    return cmd


ns = Collection("docker")
ns.configure(CONFIG)
ns.add_task(login)
ns.add_task(configure)
ns.add_task(build)
ns.add_task(buildx)
ns.add_task(push)



import pathlib
from collections.abc import Mapping
from enum import StrEnum
from typing import NotRequired, TypedDict


class UserType(StrEnum):
    AWS = "AWS"


class PodConfig(TypedDict):
    podNamespace: str
    podPrefix: str
    podPort: int
    localPort: int
    localAddress: NotRequired[str]


type PodsConfig = dict[str, PodConfig]


class PathNotFound(Exception):
    pass


class ValidationError(Exception):
    pass


def check_path(path: str) -> None:
    """Check whether a path exists on a file system."""
    if not pathlib.Path(path).exists():
        raise PathNotFound(f"Incorrect file system path: {path}")


def check_schema(config: object) -> None:
    """Check whether a port-forward configuration is valid."""
    if not isinstance(config, Mapping):
        raise ValidationError("Configuration must be an mapping")
    if len(config) == 0:
        raise ValidationError("Configuration can not be empty")
    for item in config.values():
        if not isinstance(item, Mapping):
            raise ValidationError("Configuration element must be a mapping")
        if not ({"podNamespace", "podPrefix", "podPort", "localPort"} <= set(item.keys())):
            raise ValidationError("Invalid configuration element keys")



from invoke import Collection, Context, task

from . import utils


CONFIG = {
    "helm": {
        # Sensitive
        "repo": "local-repo",
        "url": None,
        "user": "AWS",
        "token": None,
        # Non-sensitive
        "plugins": ["https://github.com/chartmuseum/helm-push.git"],
        "chart": "./helm/chart",
        "values": "./local/helm-values.yaml",
        "name": "webserver",
        "tag": "0.1.0",
    }
}


# Helm deployment workflow:


@task()
def login(
    ctx: Context,
    repo: str | None = None,
    url: str | None = None,
    user: str | None = None,
    token: str | None = None,
) -> None:
    """Login to Helm remote registry (AWS ECR or Chart Museum)."""
    repo = repo or ctx.helm.repo
    user = user or ctx.helm.user
    if user == utils.UserType.AWS:
        _aws_login(ctx, repo)
    else:
        _cm_login(ctx, user, repo, url, token)


def _aws_login(ctx: Context, repo: str) -> None:
    cmd = "aws ecr get-login-password | "
    cmd += f"helm registry login --username AWS --password-stdin {repo}"
    ctx.run(cmd)


def _cm_login(
    ctx: Context,
    user: str,
    repo: str,
    url: str | None,
    token: str | None,
) -> None:
    url = url or ctx.helm.url
    token = token or ctx.helm.token
    assert url is not None, "Missing remote Helm registry url."
    assert token is not None, "Missing remote Helm registry token."
    ctx.run(f"helm repo add {repo} {url} --username {user} --password {token}")


@task(iterable=["plugins"])
def install(ctx: Context, plugins: list[str] | None = None) -> None:
    """Provision local Helm client (Chart Museum Plugin)."""
    plugins = plugins or ctx.helm.plugins
    for plugin in plugins:
        ctx.run(f"helm plugin install {plugin}")


@task()
def lint(ctx: Context, chart=None) -> None:
    """Check compliance of Helm charts / values."""
    chart = chart or ctx.helm.chart
    utils.check_path(chart)
    ctx.run(f"helm lint {chart}")


@task(help={"dry-run": "Whether to render the Helm manifest first"})
def test(
    ctx: Context,
    chart: str | None = None,
    values: str | None = None,
    release: str = "test",
    namespace: str = "default",
    context: str = "minikube",
    dry_run: bool = False,
) -> None:
    """Test local deployment in Minikube."""
    chart = chart or ctx.helm.chart
    values = values or ctx.helm.values
    utils.check_path(chart)
    utils.check_path(values)
    cmd = f"helm install {release} {chart} --values {values} "
    cmd += f"--create-namespace --namespace {namespace} --kube-context {context}"
    if dry_run:
        cmd += " --dry-run --debug"
    ctx.run(cmd)


@task(help={"dry-run": "Whether to package the Helm chart only"})
def push(
    ctx: Context,
    chart: str | None = None,
    repo: str | None = None,
    name: str | None = None,
    tag: str | None = None,
    dry_run: bool = False,
) -> None:
    """Push Helm chart to a remote registry (AWS ECR or Chart Museum)."""
    repo = repo or ctx.helm.repo
    chart = chart or ctx.helm.chart
    utils.check_path(chart)
    if ctx.helm.user == utils.UserType.AWS:
        _aws_push(ctx, chart, repo, name, tag, dry_run)
    else:
        _cm_push(ctx, chart, repo, dry_run)


def _aws_push(
    ctx: Context,
    chart: str,
    repo: str,
    name: str | None,
    tag: str | None,
    dry_run: bool,
) -> None:
    names = (name or ctx.helm.name).split("/")
    tag = tag or ctx.helm.tag
    ctx.run(f"helm dependency update {chart}")
    ctx.run(f"helm package {chart} --version {tag}")
    if not dry_run:
        ctx.run(f"helm push {names[-1]}-{tag}.tgz oci://{repo}/{'/'.join(names[:-1])}")


def _cm_push(ctx: Context, chart: str, repo: str, dry_run: bool) -> None:
    ctx.run(f"helm dependency update {chart}")
    if not dry_run:
        ctx.run(f"helm cm-push {chart} {repo}")


ns = Collection("helm")
ns.configure(CONFIG)
ns.add_task(login)
ns.add_task(install)
ns.add_task(lint)
ns.add_task(test)
ns.add_task(push)
